"""Implement `ItemWidget`."""
from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

from kivy.app import Builder
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import (
    BooleanProperty,
    ColorProperty,
    ObjectProperty,
    StringProperty,
)

from ubo.constants import PRIMARY_COLOR

if TYPE_CHECKING:
    from . import Item


class ItemWidget(BoxLayout):
    """Renders an `Item`.

    Attributes
    ----------
    color: `tuple` of `int`
        Color of the item, used for its text and icon.

    background_color: `tuple` of `int`
        Background color of the item.

    label:
        The text rendered in the item.

    icon: `str`
        Name of a Material Symbols icon.
    """

    is_set = BooleanProperty(False)
    label = StringProperty()
    color = ColorProperty((1, 1, 1, 1))
    background_color = ColorProperty(PRIMARY_COLOR)
    icon = StringProperty()
    is_short = BooleanProperty(False)
    item = ObjectProperty()

    def on_item(self: ItemWidget, instance: ItemWidget, value: Item | None) -> None:
        if value is not None:
            instance.is_set = True
            instance.label = value['label']
            if 'is_short' in value:
                instance.is_short = value['is_short']
            if 'color' in value:
                instance.color = value['color']
            if 'background_color' in value:
                instance.background_color = value['background_color']
            if 'icon' in value:
                instance.icon = value['icon']
        else:
            instance.is_set = False


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('item_widget.kv').resolve().as_posix())



from __future__ import annotations

import pathlib
import uuid
import warnings
from datetime import datetime, timezone
from enum import Enum, auto
from typing import TYPE_CHECKING, Callable

from kivy.app import Builder
from kivy.event import EventDispatcher
from kivy.properties import ColorProperty, ObjectProperty, StringProperty

from ubo.page import PAGE_MAX_ITEMS, PageWidget

if TYPE_CHECKING:
    from typing_extensions import Any

    from ubo.menu.types import Item


class Importance(Enum):
    CRITICAL = auto()
    HIGH = auto()
    MEDIUM = auto()
    LOW = auto()


IMPORTANCE_ICONS = {
    Importance.CRITICAL: 'dangerous',
    Importance.HIGH: 'warning',
    Importance.MEDIUM: 'info',
    Importance.LOW: 'priority',
}

IMPORTANCE_COLORS = {
    Importance.CRITICAL: '#D32F2F',
    Importance.HIGH: '#FFA000',
    Importance.MEDIUM: '#FFEB3B',
    Importance.LOW: '#2196F3',
}


class NotificationAction:
    title: str
    callback: Callable

    def __init__(self: NotificationAction, *, title: str, callback: Callable) -> None:
        self.title = title
        self.callback = callback

    def __repr__(self: NotificationAction) -> str:
        return f'<NotificationAction(title={self.title}, callback={self.callback})>'


class Notification:
    title: str
    content: str
    importance: Importance
    timestamp: datetime
    is_read: bool
    sender: str | None
    actions: list[NotificationAction]
    icon: str
    expiry_date: datetime | None

    def __init__(self: Notification, *,
                 title: str,
                 content: str,
                 importance: Importance = Importance.LOW,
                 sender: str | None = None,
                 actions: list[NotificationAction] | None = None,
                 icon: str | None = None,
                 expiry_date: datetime | None = None,
                 timestamp: datetime | None = None,
                 ) -> None:
        if actions is None:
            actions = []
        if timestamp is None:
            timestamp = datetime.now(tz=timezone.utc)
        if icon is None:
            icon = IMPORTANCE_ICONS[importance]

        self.id = str(uuid.uuid4())
        self.title = title
        self.content = content
        self.importance = importance
        self.timestamp = timestamp
        self.is_read = False
        self.sender = sender
        self.actions = actions
        self.icon = icon
        self.expiry_date = expiry_date

    def mark_read(self: Notification) -> None:
        self.is_read = True

    def mark_unread(self: Notification) -> None:
        self.is_read = False

    def is_expired(self: Notification) -> bool:
        return self.expiry_date is not None and\
            datetime.now(tz=timezone.utc) > self.expiry_date

    def __repr__(self) -> str:
        return (f'<Notification(id={self.id}, title={self.title}, '
                f'content={self.content}, importance={self.importance.name}, '
                f'timestamp={self.timestamp}, is_read={self.is_read}, '
                f'sender={self.sender}, actions={self.actions}, icon={self.icon}, '
                f'expiry_date={self.expiry_date})>')


class NotificationManager(EventDispatcher):
    __events__ = ('on_change',)

    _notifications: list[Notification] = []

    def notify(self: NotificationManager, *,
               title: str,
               content: str,
               importance: Importance = Importance.LOW,
               sender: str | None = None,
               actions: list[NotificationAction] | None = None,
               icon: str | None = None,
               expiry_date: datetime | None = None) -> None:

        self._notifications.append(Notification(
            title=title,
            content=content,
            importance=importance,
            sender=sender,
            actions=actions,
            icon=icon,
            expiry_date=expiry_date,
        ))
        self.dispatch('on_change')

    @property
    def unread_count(self: NotificationManager):
        return len([
            notification
            for notification in self._notifications
            if not notification.is_read
        ])

    def remove(self: NotificationManager, notification: Notification):
        self._notifications.remove(notification)
        self.dispatch('on_change')

    def menu_items(self: NotificationManager) -> list[Item]:
        def notification_widget_builder(
            notification: Notification,
            index: int,
        ) -> Callable[[], NotificationWidget]:
            return lambda **kwargs: NotificationWidget(
                notification=notification,
                title=f'Notification ({index+1}/'
                f'{len(self._notifications)})',
                **kwargs,
            )

        return [{
            'label': notification.title,
            'icon': notification.icon,
            'color': 'black',
            'background_color': IMPORTANCE_COLORS[notification.importance],
            'application': notification_widget_builder(notification, index),
        } for index, notification in enumerate(self._notifications)]

    def on_change(self: NotificationManager) -> None:
        pass


notification_manager = NotificationManager()


class NotificationWidget(PageWidget):
    notification = ObjectProperty()
    color = ColorProperty()
    title = StringProperty()

    def __init__(
        self: NotificationWidget, *,
        notification: Notification,
        title: str,
        **kwargs: Any,
    ) -> None:
        self.notification = notification
        self.color = IMPORTANCE_COLORS[notification.importance]
        self.title = title
        super().__init__(items=[{
            'icon': 'delete',
            'action': lambda: (
                notification_manager.remove(notification),
                self.dispatch('on_close'),
            ),
            'label': '',
            'is_short': True,
        }], **kwargs)

    def get_item(self: NotificationWidget, index: int) -> Item | None:
        if index != PAGE_MAX_ITEMS - 1:
            warnings.warn('index must be 2', ResourceWarning, stacklevel=1)
            return None
        return self.items[index - 2]


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('notification_widget.kv').resolve().as_posix())



from __future__ import annotations

import pathlib
import warnings
from typing import TYPE_CHECKING, Any

from kivy.app import Builder
from kivy.properties import ObjectProperty, StringProperty

from ubo.page import PageWidget

if TYPE_CHECKING:
    from ubo.menu.types import Item

PROMPT_OPTIONS = 2


class PromptWidget(PageWidget):
    icon = StringProperty()
    prompt = StringProperty()
    first_option_label = StringProperty()
    first_option_icon = StringProperty()
    first_option_callback = ObjectProperty()
    second_option_label = StringProperty()
    second_option_icon = StringProperty()
    second_option_callback = ObjectProperty()

    def __init__(
        self: PromptWidget,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        super().__init__(
            items=[
                {
                    'label': self.first_option_label,
                    'icon': self.first_option_icon,
                    'action': self.first_option_callback,
                },
                {
                    'label': self.second_option_label,
                    'icon': self.second_option_icon,
                    'action': self.second_option_callback,
                },
            ],
            **kwargs,
        )

    def get_item(self: PromptWidget, index: int) -> Item | None:
        if not 1 <= index <= PROMPT_OPTIONS:
            warnings.warn('index must be either 1 or 2',
                          ResourceWarning, stacklevel=1)
            return None
        return self.items[index - 1]


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('prompt_widget.kv').resolve().as_posix())



from __future__ import annotations

from kivy.animation import Animation
from kivy.properties import NumericProperty
from kivy.uix.slider import Slider


class AnimatedSlider(Slider):
    animated_value = NumericProperty(0)

    def __init__(self, **kwargs):
        self.value = self.animated_value
        super().__init__(**kwargs)

    def on_animated_value(
        self: AnimatedSlider,
        instance: AnimatedSlider,
        new_value: float,
    ):
        Animation(value=new_value, duration=0.2).start(self)



import pathlib

from kivy.app import Builder

from ubo.page import PageWidget


class NormalMenuPageWidget(PageWidget):
    """renders a normal page of a `Menu`."""


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('normal_menu_page_widget.kv').resolve().as_posix())



from __future__ import annotations

import pathlib
import warnings
from typing import TYPE_CHECKING, Any

from kivy.app import Builder, StringProperty

from ubo.page import PageWidget

from .constants import PAGE_SIZE

if TYPE_CHECKING:
    from .types import Item


class HeaderMenuPageWidget(PageWidget):
    """renders a header page of a `Menu`."""

    heading = StringProperty()
    sub_heading = StringProperty()

    def __init__(
        self: HeaderMenuPageWidget,
        item: Item,
        heading: str,
        sub_heading: str,
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize a `HeaderMenuPageWidget`.

        Parameters
        ----------
        item: `Item`
            The item to be shown in this page

        heading: `str`
            The heading of the page

        sub_heading: `str`
            The sub-heading of the page

        kwargs: Any
            Stuff that will get directly passed to the `__init__` method of Kivy's
        `Screen`.
        """
        super().__init__([item], **kwargs)
        self.heading = heading
        self.sub_heading = sub_heading

    def get_item(self: HeaderMenuPageWidget, index: int) -> Item | None:
        if index != PAGE_SIZE - 1:
            warnings.warn('index must be 2', ResourceWarning, stacklevel=1)
            return None
        return self.items[index - 2]


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('header_menu_page_widget.kv').resolve().as_posix())



from __future__ import annotations

import logging
import math
import time
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypedDict

import adafruit_aw9523
from adafruit_bus_device import i2c_device

IS_RPI = Path('/etc/rpi-issue').exists()
if IS_RPI:
    import board
    from RPi import GPIO

if TYPE_CHECKING:
    from adafruit_register.i2c_struct import UnaryStruct

INT_EXPANDER = 5  # GPIO PIN index that receives interrupt from AW9523


class KeypadError(Exception):
    ...


class ButtonName(Enum):
    TOP_LEFT = 'top_left'
    MIDDLE_LEFT = 'middle_left'
    BOTTOM_LEFT = 'bottom_left'
    UP = 'up'
    DOWN = 'down'
    BACK = 'back'
    HOME = 'home'
    MIC = 'mic'


ButtonStatus = Literal['pressed', 'released']


class ButtonEvent(TypedDict):
    status: ButtonStatus
    timestamp: float


class Event(TypedDict):
    inputs: UnaryStruct
    timestamp: float


class KeypadStatus:
    """Class to keep track of button status.

    0: "TOP_LEFT": Top left button
    1: "MIDDLE_LEFT": Middle left button
    2: "BOTTOM_LEFT": Bottom left button
    3: "BACK": Button with back arrow label (under LCD)
    4: "HOME": Button with home label (Under LCD)
    5: "UP": Right top button with upward arrow label
    6: "DOWN": Right bottom button with downward arrow label
    7: "MIC": Microphone mute switch
    """

    buttons: dict[ButtonName, ButtonEvent]

    def __init__(self: KeypadStatus) -> None:
        self.buttons = {
            button_name: {'status': 'released', 'timestamp': time.time()}
            for button_name in ButtonName
        }

    def update_status(self: KeypadStatus,
                      button_name: ButtonName,
                      new_status: Literal['pressed', 'released']) -> None:
        if new_status not in {'pressed', 'released'}:
            msg = 'Invalid status'
            raise KeypadError(msg)
        if button_name in self.buttons:
            self.buttons[button_name]['status'] = new_status
            self.buttons[button_name]['timestamp'] = time.time()

    def get_status(self: KeypadStatus,
                   button_name: ButtonName) -> Literal['pressed', 'released']:
        if button_name in self.buttons:
            return self.buttons[button_name]['status']
        msg = 'Invalid button name'
        raise KeypadError(msg)

    def get_timestamp(self: KeypadStatus, button_name: ButtonName) -> float:
        if button_name in self.buttons:
            return self.buttons[button_name]['timestamp']
        msg = 'Invalid button name'
        raise KeypadError(msg)

    def get_label(self: KeypadStatus, index: int) -> ButtonName:
        if index > len(ButtonName):
            msg = 'Invalid index'
            raise KeypadError(msg)
        return list(ButtonName)[index]


class Keypad:
    """Class to handle keypad events."""

    event_queue: list[Event]
    aw: adafruit_aw9523.AW9523 | None
    inputs: UnaryStruct | None

    def __init__(self: Keypad) -> None:
        """Initialize a Keypad.

        Initializes various parameters including
        loggers and button names.
        """
        self.logger = logging.getLogger('keypad')
        self.logger.debug('Initialising keypad...')
        self.event_queue = []
        self.aw = None
        self.inputs = None
        self.bus_address = 0x58
        self.model = 'aw9523'
        self.enabled = True
        self.buttons = KeypadStatus()
        self.index = 0
        if IS_RPI:
            self.init_i2c()

    def clear_interrupt_flags(self: Keypad, i2c: i2c_device.I2CDevice) -> None:
        # Write to both registers to reset the interrupt flag
        buffer = bytearray(2)
        buffer[0] = 0x00
        buffer[1] = 0x00
        i2c.write(buffer)
        i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)

        time.sleep(0.1)
        buffer[0] = 0x01
        buffer[1] = 0x00
        i2c.write(buffer)
        i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)
        time.sleep(0.1)

    def disable_interrupt_for_higher_bits(self: Keypad,
                                          i2c: i2c_device.I2CDevice) -> None:
        # disable interrupt for higher bits
        buffer = bytearray(2)
        buffer[0] = 0x06
        buffer[1] = 0x00
        i2c.write(buffer)
        i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)

        buffer[0] = 0x07
        buffer[1] = 0xff
        i2c.write(buffer)
        i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)

    def init_i2c(self: Keypad) -> None:
        # connect to the I2C bus
        GPIO.setmode(GPIO.BCM)
        i2c = board.I2C()
        # Set this to the GPIO of the interrupt:
        GPIO.setup(INT_EXPANDER, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        try:
            # Search for the GPIO expander address on the I2C bus
            self.aw = adafruit_aw9523.AW9523(i2c, self.bus_address)
            new_i2c = self.aw.i2c_device
        except Exception:
            self.bus_address = False
            self.logger.exception(
                'Failed to initialize I2C Bus on address 0x58')
            raise

        # Perform soft reset of the expander
        self.aw.reset()
        # Set first 8 low significant bits (register 1) to input
        self.aw.directions = 0xff00
        time.sleep(1)

        # The code below, accessing the GPIO expander registers directly via i2c
        # was created as a workaround of the reset the interrupt flag.
        self.clear_interrupt_flags(new_i2c)
        self.disable_interrupt_for_higher_bits(new_i2c)
        # reset interrupts again
        self.clear_interrupt_flags(new_i2c)

        # read register values
        self.inputs = self.aw.inputs
        self.logger.debug('Initializing inputs', extra={
                          'inputs': f'{self.inputs:016b}'})
        self.event_queue = [{'inputs': self.inputs, 'timestamp': time.time()}]
        time.sleep(0.5)

        # Enable interrupt on the GPIO expander
        GPIO.add_event_detect(INT_EXPANDER,
                              GPIO.FALLING,
                              callback=self.key_press_cb,
                              bouncetime=1)

    def key_press_cb(self: Keypad, _channel):
        """Handle key press dispatched by GPIO interrupt.

            This is callback function that gets triggers
         if any change is detected on keypad buttons
         states.

         In this callback, we look at the state
         change to see which button was pressed or leased.

        Parameters
        ----------
        _channel: int
            GPIO channel that triggered the callback
            NOT USED currently
        """
        if self.aw is None:
            return
        # read register values
        self.inputs = self.aw.inputs
        event: Event = {'inputs': self.inputs, 'timestamp': time.time()}
        # append the event to the queue. The queue has a depth of 2 and
        # keeps the current and last event.
        self.event_queue.append(event)
        self.logger.info(self.event_queue)
        self.logger.debug('Current Inputs', extra={
                          'inputs': f'{self.inputs:016b}'})
        previos_event = self.event_queue.pop(0)
        self.logger.debug('Previous Inputs', extra={
                          'inputs': f'{previos_event.get("inputs"):016b}'})
        # XOR the last recorded input values with the current input values
        # to see which bits have changed. Technically there can only be one
        # bit change in every callback
        change_mask = previos_event.get('inputs') ^ self.inputs
        self.logger.debug('Change', extra={
                          'change_mask': f'{change_mask:016b}'})
        # use the change mask to see if the button was the change was
        # falling (1->0) indicating a pressed action
        # or risign (0->1) indicating a release action
        self.index = (int)(math.log2(change_mask))
        self.logger.info('button index', extra={'button_index': self.index})
        # Check for rising edge or falling edge action (press or release)
        self.button_label = self.buttons.get_label(self.index)
        if ((previos_event.get('inputs') & change_mask) == 0):
            self.logger.info(
                'Button pressed',
                extra={'button': str(self.index), 'label': self.button_label},
            )

            # calculate how long the button was held down
            # and print the time
            last_time_stamp = self.buttons.get_timestamp(self.button_label)
            held_down_time = time.time() - last_time_stamp
            self.logger.info(
                'Button was pressed down',
                extra={'held_down_time': held_down_time},
            )

            self.buttons.update_status(self.button_label, 'released')
            self.on_button_event(self.button_label, 'released')

        else:
            self.logger.info(
                'Button released',
                extra={'button': str(self.index), 'label': self.button_label},
            )
            self.buttons.update_status(self.button_label, 'pressed')
            self.on_button_event(self.button_label, 'pressed')

        self.logger.info(self.buttons.buttons)

    def on_button_event(
        self: Keypad,
        button_name: ButtonName,
        button_status: ButtonStatus,
    ) -> None:
        raise NotImplementedError



"""Demo of ubo-menu widget."""
from __future__ import annotations

import datetime
import os
from functools import cached_property
from threading import Thread
from typing import TYPE_CHECKING, Literal

from headless_kivy_pi import setup_headless
from kivy.app import Widget
from kivy.base import Clock
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label

from ubo.gauge import GaugeWidget
from ubo.keypad import ButtonName, ButtonStatus, Keypad
from ubo.volume import VolumeWidget

os.environ['KIVY_METRICS_DENSITY'] = '1'
os.environ['KIVY_NO_CONFIG'] = '1'
os.environ['KIVY_NO_FILELOG'] = '1'

setup_headless()

from kivy.core.window import (  # noqa: E402
    Keyboard,
    Window,
    WindowBase,
)

from ubo.app import UboApp  # noqa: E402
from ubo.menu import MenuWidget  # noqa: E402
from ubo.notification import (  # noqa: E402
    Importance,
    notification_manager,
)
from ubo.prompt import PromptWidget  # noqa: E402

if TYPE_CHECKING:
    from menu import Menu
    Modifier = Literal['ctrl', 'alt', 'meta', 'shift']

notification_manager.notify(
    title='Low priority',
    content='Something happened but it is not important and this content is very long'
    'since we need a very long content to check the scrollability of the widget',
    importance=Importance.LOW,
    sender='demo',
)
notification_manager.notify(
    title='Medium priority',
    content='Something happened and it is somehow important',
    importance=Importance.MEDIUM,
    sender='demo',
)
notification_manager.notify(
    title='High priority',
    content='Something happened and it is important',
    importance=Importance.HIGH,
    sender='demo',
)
notification_manager.notify(
    title='Critical priority',
    content='Something happened and it is critically important',
    importance=Importance.CRITICAL,
    sender='demo',
)


class WifiPrompt(PromptWidget):
    icon = 'wifi_off'
    prompt = 'Not Connected'
    first_option_label = 'Add'
    first_option_icon = 'add'

    def first_option_callback(self: WifiPrompt):
        notification_manager.notify(
            title='Wifi added',
            content='This Wifi network has been added',
            icon='wifi',
            sender='Wifi',
        )

    second_option_label = 'Forget'
    second_option_icon = 'delete'

    def second_option_callback(self: WifiPrompt):
        notification_manager.notify(
            title='Wifi forgot',
            content='This Wifi network is forgotten',
            importance=Importance.CRITICAL,
            sender='Wifi',
        )


SETTINGS_MENU: Menu = {
    'title': 'Settings',
    'heading': 'Please choose',
    'sub_heading': 'This is sub heading',
    'items': [
        {
            'label': 'WiFi',
            'application': WifiPrompt,
            'icon': 'wifi',
        },
        {
            'label': 'Bluetooth',
            'action': lambda: print('Bluetooth'),
            'icon': 'bluetooth',
        },
        {
            'label': 'Audio',
            'action': lambda: print('Audio'),
            'icon': 'volume_up',
        },
    ],
}

MAIN_MENU: Menu = {
    'title': 'Main',
    'heading': 'What are you going to do?',
    'sub_heading': 'Choose from the options',
    'items': [
        {
            'label': 'Settings',
            'icon': 'settings',
            'sub_menu': SETTINGS_MENU,
        },
        {
            'label': 'Apps',
            'action': lambda: print('Apps'),
            'icon': 'apps',
        },
        {
            'label': 'About',
            'action': lambda: print('About'),
            'icon': 'info',
        },
    ],
}


HOME_MENU: Menu = {
    'title': 'Dashboard',
    'items': [
        {
            'label': '',
            'sub_menu': MAIN_MENU,
            'icon': 'menu',
            'is_short': True,
        },
        {
            'label': '',
            'sub_menu': {
                'title': lambda: f'Notifications ({notification_manager.unread_count})',
                'items': notification_manager.menu_items,
            },
            'color': 'yellow',
            'icon': 'info',
            'is_short': True,
        },
        {
            'label': 'Turn off',
            'action': lambda: print('Turning off'),
            'icon': 'power_settings_new',
            'is_short': True,
        },
    ],
}


class MenuApp(UboApp, Keypad):
    """Menu application."""

    def build(self: MenuApp) -> Widget | None:
        Window.bind(on_keyboard=self.on_keyboard)
        return super().build()

    @cached_property
    def menu_widget(self: MenuApp) -> MenuWidget:
        """Build the main menu and initiate it."""
        menu_widget = MenuWidget()
        menu_widget.set_current_menu(HOME_MENU)

        def title_callback(_: MenuWidget, title: str):
            self.root.title = title
        self.root.title = menu_widget.title
        menu_widget.bind(title=title_callback)

        return menu_widget

    @cached_property
    def cpu_gauge(self: MenuApp) -> GaugeWidget:
        import psutil
        gauge = GaugeWidget(value=0, fill_color='#24D636', label='CPU')

        value = [0]

        def set_value(_dt: float) -> None:
            gauge.value = value[0]

        def calculate_value() -> None:
            value[0] = psutil.cpu_percent(interval=1, percpu=False)
            Clock.schedule_once(set_value)

        Clock.schedule_interval(
            lambda _dt: Thread(target=calculate_value).start(),
            1,
        )

        return gauge

    @cached_property
    def ram_gauge(self: MenuApp) -> GaugeWidget:
        import psutil
        gauge = GaugeWidget(
            value=psutil.virtual_memory().percent,
            fill_color='#D68F24',
            label='RAM',
        )

        def set_value(_dt: int) -> None:
            gauge.value = psutil.virtual_memory().percent

        Clock.schedule_interval(set_value, 1)

        return gauge

    @cached_property
    def clock_widget(self: MenuApp) -> Label:
        clock = Label(font_size=dp(20))
        local_timzone = datetime.datetime.now(
            datetime.timezone.utc).astimezone().tzinfo

        def now() -> datetime.datetime:
            return datetime.datetime.now(local_timzone)

        def set_value(_dt: int) -> None:
            clock.text = now().strftime('%H:%M')

        set_value(0)

        def initialize(_dt: int) -> None:
            Clock.schedule_interval(set_value, 60)

        Clock.schedule_once(initialize, 60 - now().second + 1)

        return clock

    @cached_property
    def central(self: MenuApp) -> Widget:
        horizontal_layout = BoxLayout()

        self.menu_widget.size_hint = (None, 1)
        self.menu_widget.width = dp(50)
        horizontal_layout.add_widget(self.menu_widget)

        central_column = BoxLayout(
            orientation='vertical', spacing=dp(12), padding=dp(16))
        central_column.add_widget(self.cpu_gauge)
        central_column.add_widget(self.ram_gauge)
        central_column.size_hint = (1, 1)
        horizontal_layout.add_widget(central_column)

        right_column = BoxLayout(orientation='vertical')
        right_column.add_widget(VolumeWidget(value=40))
        right_column.size_hint = (None, 1)
        horizontal_layout.add_widget(right_column)

        def handle_depth_change(_instance: Widget, depth: int) -> None:
            if depth == 0:
                self.menu_widget.size_hint = (None, 1)
                self.menu_widget.width = dp(50)
                central_column.size_hint = (1, 1)
                right_column.size_hint = (None, 1)
            else:
                self.menu_widget.size_hint = (1, 1)
                central_column.size_hint = (0, 1)
                right_column.size_hint = (0, 1)

        self.menu_widget.bind(depth=handle_depth_change)

        return horizontal_layout

    @cached_property
    def footer(self: MenuApp) -> Widget:
        layout = BoxLayout()

        normal_footer_layout = BoxLayout(
            orientation='horizontal', spacing=0, padding=0)
        normal_footer_layout.add_widget(Label(
            text='reply',
            font_name='material_symbols',
            font_size=dp(20),
            font_features='fill=1',
            size_hint=(None, 1),
        ))
        normal_footer_layout.add_widget(Widget(size_hint=(1, 1)))

        home_footer_layout = BoxLayout(
            orientation='horizontal', spacing=0, padding=0)
        home_footer_layout.add_widget(
            Widget(size_hint=(None, 1), width=dp(16)))
        home_footer_layout.add_widget(self.clock_widget)
        home_footer_layout.add_widget(Widget(size_hint=(1, 1)))

        for icon in ['camera', 'lan', 'mic_off', 'bluetooth', 'wifi_off']:
            label = Label(
                text=icon,
                font_name='material_symbols',
                font_size=dp(20),
                font_features='fill=1',
                size_hint=(.5, 1),
            )
            home_footer_layout.add_widget(label)
            if icon == 'camera':
                label.color = 'green'

        home_footer_layout.add_widget(
            Widget(size_hint=(None, 1), width=dp(16)))

        def handle_depth_change(_instance: Widget, depth: int) -> None:
            if depth == 0:
                if normal_footer_layout in layout.children:
                    layout.remove_widget(normal_footer_layout)
                    layout.add_widget(home_footer_layout)
            elif home_footer_layout in layout.children:
                layout.remove_widget(home_footer_layout)
                layout.add_widget(normal_footer_layout)

        self.menu_widget.bind(depth=handle_depth_change)

        layout.add_widget(home_footer_layout)

        return layout

    def on_keyboard(
        self: MenuApp,
        _: WindowBase,
        key: int,
        _scancode: int,
        _codepoint: str,
        modifier: list[Modifier],
    ) -> None:
        """Handle keyboard events."""
        if modifier == []:
            if key == Keyboard.keycodes['up']:
                self.menu_widget.go_to_previous_page()
            elif key == Keyboard.keycodes['down']:
                self.menu_widget.go_to_next_page()
            elif key == Keyboard.keycodes['1']:
                self.menu_widget.select(0)
            elif key == Keyboard.keycodes['2']:
                self.menu_widget.select(1)
            elif key == Keyboard.keycodes['3']:
                self.menu_widget.select(2)
            elif key == Keyboard.keycodes['left']:
                self.menu_widget.go_back()

    def on_button_event(
        self: MenuApp,
        button_pressed: ButtonName,
        button_status: ButtonStatus,
    ) -> None:
        if button_status == 'pressed':
            if button_pressed == ButtonName.UP:
                Clock.schedule_once(
                    lambda dt: self.menu_widget.go_to_previous_page(), -1)
            elif button_pressed == ButtonName.DOWN:
                Clock.schedule_once(
                    lambda dt: self.menu_widget.go_to_next_page(), -1)
            elif button_pressed == ButtonName.TOP_LEFT:
                Clock.schedule_once(lambda dt: self.menu_widget.select(0), -1)
            elif button_pressed == ButtonName.MIDDLE_LEFT:
                Clock.schedule_once(lambda dt: self.menu_widget.select(1), -1)
            elif button_pressed == ButtonName.BOTTOM_LEFT:
                Clock.schedule_once(lambda dt: self.menu_widget.select(2), -1)
            elif button_pressed == ButtonName.BACK:
                Clock.schedule_once(lambda dt: self.menu_widget.go_back(), -1)
        self.root.reset_fps_control_queue()


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    app = MenuApp()
    app.run()


if __name__ == '__main__':
    main()



"""Implement a paginated menu.

The first page starts with a heading and its sub-heading.
Each item may have sub items, in that case activating this item will open a new menu
with its sub items.
Each item can optionally be styled differently.
"""
from __future__ import annotations

import pathlib
import uuid
import warnings
from typing import TYPE_CHECKING, cast

from headless_kivy_pi import HeadlessWidget
from kivy.app import Builder
from kivy.properties import AliasProperty, NumericProperty, StringProperty
from kivy.uix.boxlayout import BoxLayout

from .constants import PAGE_SIZE
from .header_menu_page_widget import HeaderMenuPageWidget
from .item_widget import ItemWidget  # noqa: F401
from .normal_menu_page_widget import NormalMenuPageWidget
from .types import (
    Item,
    is_action_item,
    is_application_item,
    is_sub_menu_item,
    menu_items,
    menu_title,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from kivy.uix.screenmanager import ScreenManager
    from menu.types import Menu
    from page import PageWidget
    from typing_extensions import Any

    from ubo.animated_slider import AnimatedSlider


def paginate(items: list[Item], offset: int = 0) -> Iterator[list[Item]]:
    """Yield successive PAGE_SIZE-sized chunks from list.

    Parameters
    ----------
    items: `list` of `Item`
        The items to be paginated.

    offset: `int`
        An offset greater than or equal to zero and less than `PAGE_SIZE`. The size of
        the first page will be `PAGE_SIZE` - `offset`. The default value is 0.
    """
    for i in range(PAGE_SIZE-offset, len(items), PAGE_SIZE):
        yield items[i:i + PAGE_SIZE]


class MenuWidget(BoxLayout):
    """Paginated menu."""

    def get_pages(self: MenuWidget) -> list[PageWidget]:
        return self._pages

    def set_pages(self: MenuWidget, pages: list[PageWidget]) -> None:
        self._pages = pages

    title = StringProperty(allownone=True)
    page_index = NumericProperty(0)
    depth = NumericProperty(0)
    pages = AliasProperty(getter=get_pages, setter=set_pages, bind=['depth'])
    _pages: list[PageWidget]
    current_menu: Menu = None
    current_application: PageWidget | None = None
    menu_stack: list[tuple[Menu, int]]
    screen_manager: ScreenManager
    slider: AnimatedSlider

    def __init__(self: MenuWidget, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize a `MenuWidget`."""
        self.pages = []
        self.menu_stack = []
        super().__init__(**kwargs)
        HeadlessWidget.activate_low_fps_mode()

    @property
    def current_depth(self: MenuWidget) -> int:
        """Depth of current menu in menu tree."""
        return len(self.menu_stack) + (1 if self.current_application else 0)

    @property
    def current_menu_items(self: MenuWidget) -> list[Item]:
        """Items of the current menu."""
        if self.current_application:
            return []
        return menu_items(self.current_menu)

    def go_to_next_page(self: MenuWidget) -> None:
        """Go to the next page.

        If it is already the last page, rotate to the first page.
        """
        if len(self.current_menu_items) == 0:
            return
        self.page_index += 1
        if self.page_index >= len(self.pages):
            self.page_index = 0
        self.screen_manager.transition.direction = 'up'
        self.update()

    def go_to_previous_page(self: MenuWidget) -> None:
        """Go to the previous page.

        If it is already the first page, rotate to the last page.
        """
        if len(self.current_menu_items) == 0:
            return
        self.page_index -= 1
        if self.page_index < 0:
            self.page_index = len(self.pages) - 1
        self.screen_manager.transition.direction = 'down'
        self.update()

    def select(self: MenuWidget, index: int) -> None:
        """Select one of the items currently visible on the screen.

        Parameters
        ----------
        index: `int`
            An integer number, can only take values greater than or equal to zero and
            less than `PAGE_SIZE`
        """
        if self.screen_manager.current_screen is None:
            warnings.warn('`current_screen` is `None`',
                          RuntimeWarning, stacklevel=1)
            return
        current_page: PageWidget = self.screen_manager.current_screen
        item = current_page.get_item(index)
        if not item:
            warnings.warn('Selected `item` is `None`',
                          RuntimeWarning, stacklevel=1)
            return
        if is_action_item(item):
            item['action']()
        elif is_sub_menu_item(item):
            self.push_menu(item['sub_menu'])
        elif is_application_item(item):
            self.open_application(item['application'](name=uuid.uuid4().hex))

    def open_application(self: MenuWidget, application: PageWidget) -> None:
        """Open an application."""
        HeadlessWidget.activate_high_fps_mode()
        self.current_application = application
        self.pages.append(self.current_application)
        self.screen_manager.add_widget(self.current_application)
        self.screen_manager.transition.direction = 'left'
        self.screen_manager.current = self.current_application.name
        self.title = self.current_application.title if hasattr(
            self.current_application, 'title') else None
        self.current_application.bind(on_close=lambda _: self.go_back())
        self.depth = self.current_depth

    def go_back(self: MenuWidget) -> None:
        """Go back to the previous menu."""
        self.screen_manager.transition.direction = 'right'
        if self.current_application:
            HeadlessWidget.activate_high_fps_mode()
            self.current_application = None
            self.set_current_menu(self.current_menu)
        else:
            self.pop_menu()
        self.depth = self.current_depth

    def update(self: MenuWidget) -> None:
        """Activate the transition from the previously active page to the current page.

        Activate high fps mode to render the animation in high fps
        """
        HeadlessWidget.activate_high_fps_mode()
        self.screen_manager.current = f'Page {self.current_depth} {self.page_index}'

    def push_menu(self: MenuWidget, menu: Menu) -> None:
        """Go one level deeper in the menu stack."""
        self.screen_manager.transition.direction = 'left'
        self.menu_stack.append((self.current_menu, self.page_index))
        self.set_current_menu(menu)
        self.depth = self.current_depth

    def pop_menu(self: MenuWidget) -> None:
        """Come up one level from of the menu stack."""
        if self.current_depth == 0:
            return
        parent_menu = self.menu_stack.pop()
        self.set_current_menu(parent_menu[0])
        self.page_index = parent_menu[1]
        self.screen_manager.current = f'Page {self.current_depth} {self.page_index}'
        self.depth = self.current_depth

    def set_current_menu(self: MenuWidget, menu: Menu) -> None:
        """Set the `current_menu` and create its pages."""
        HeadlessWidget.activate_high_fps_mode()
        while len(self.pages) > 0:
            self.screen_manager.remove_widget(self.pages.pop())

        self.page_index = 0
        self.current_menu = menu

        if 'heading' in self.current_menu:
            first_page = HeaderMenuPageWidget(
                menu_items(menu)[0],
                cast(str, menu.get('heading', '')),
                cast(str, menu.get('sub_heading', '')),
                name=f'Page {self.current_depth} 0',
            )
        else:
            first_page = NormalMenuPageWidget(
                menu_items(menu)[:3], name=f'Page {self.current_depth} 0')
        self.pages.append(first_page)
        self.screen_manager.add_widget(first_page)

        paginated_items = paginate(
            menu_items(menu), 2 if 'heading' in menu else 0)
        for index, page_items in enumerate(paginated_items):
            page = NormalMenuPageWidget(
                page_items, name=f'Page {self.current_depth} {index + 1}')
            self.pages.append(page)
            self.screen_manager.add_widget(page)
        self.title = menu_title(menu)
        self.screen_manager.current = f'Page {self.current_depth} 0'
        self.slider.value = len(self.pages) - 1 - self.page_index
        HeadlessWidget.activate_low_fps_mode()

    def on_kv_post(self: MenuWidget, _: Any) -> None:  # noqa: ANN401
        """Activate low fps mode when transition is done."""
        self.screen_manager = self.ids.screen_manager
        on_progress_ = self.screen_manager.transition.on_progress

        def on_progress(progression):
            self.screen_manager.transition.screen_out.opacity = (
                1 - progression)
            self.screen_manager.transition.screen_in.opacity = progression
            on_progress_(progression)

        def on_complete():
            self.screen_manager.transition.screen_out.opacity = 0
            self.screen_manager.transition.screen_in.opacity = 1
            HeadlessWidget.activate_low_fps_mode()

        self.screen_manager.transition.on_progress = on_progress
        self.screen_manager.transition.on_complete = on_complete

        self.slider = self.ids.slider


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('menu.kv').resolve().as_posix())



from __future__ import annotations

import pathlib

from kivy.app import Builder
from kivy.properties import ColorProperty, NumericProperty
from kivy.uix.boxlayout import BoxLayout

from ubo.constants import PRIMARY_COLOR, SECONDARY_COLOR


class VolumeWidget(BoxLayout):
    value = NumericProperty(50)
    fill_color = ColorProperty(PRIMARY_COLOR)
    background_color = ColorProperty(SECONDARY_COLOR)


Builder.load_file(pathlib.Path(
    __file__).parent.joinpath('volume_widget.kv').resolve().as_posix())



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from kivy.core.window import ListProperty
from kivy.uix.screenmanager import Screen

if TYPE_CHECKING:
    from menu.types import Item
    from typing_extensions import Any


PAGE_MAX_ITEMS = 3


class PageWidget(Screen):
    """renders a page."""

    __events__ = ('on_close', )

    items = ListProperty([])

    def __init__(
        self: PageWidget,
        items: list[Item],
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Initialize a `PageWidget`.

        Parameters
        ----------
        items: `list` of `Item`
            The items to be shown in this page

        kwargs: Any
            Stuff that will get directly passed to the `__init__` method of Kivy's
        `Screen`.
        """
        if len(items) > PAGE_MAX_ITEMS:
            msg = f'`Page` is initialized with more than `MAX_ITEMS`={PAGE_MAX_ITEMS} items'
            raise ValueError(msg)
        self.items = items
        super().__init__(**kwargs)

    def get_item(self: PageWidget, index: int) -> Item | None:
        if not 0 <= index < len(self.items):
            msg = f"""index must be greater than or equal to 0 and less than {
            len(self.items)}"""
            warnings.warn(msg, ResourceWarning, stacklevel=1)
            return None
        return self.items[index]

    def on_close(self: PageWidget) -> None:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Union

from typing_extensions import NotRequired, TypedDict, TypeGuard

if TYPE_CHECKING:
    from page import PageWidget


class BaseMenu(TypedDict):
    """A class used to represent a menu.

    Attributes
    ----------
    title: `str`
        Rendered on top of the menu in all pages. Optionally it can be a callable
    returning the title.

    items: `list` of `Item`
        List of the items of the menu. Optionally it can be a callable returning the
    list of items
    """

    title: str | Callable[[], str]
    items: list[Item] | Callable[[], list[Item]]


class HeadedMenu(BaseMenu):
    """A class used to represent a headed menu.

    Attributes
    ----------
    heading: `str`
        Rendered in the first page of the menu, stating the purpose of the menu and its
    items.

    sub_heading: `str`
        Rendered beneath the heading in the first page of the menu with a smaller font.
    """

    heading: str
    sub_heading: str


def is_headed_menu(menu: Menu) -> TypeGuard[ActionItem]:
    """Check whether the menu is a `HeadedMenu` or not."""
    return 'heading' in menu and 'sub_heading' in menu


class HeadlessMenu(BaseMenu):
    """A class used to represent a headless menu."""


def is_headless_menu(menu: Menu) -> TypeGuard[ActionItem]:
    """Check whether the menu is a `HeadedMenu` or not."""
    return 'heading' not in menu and 'sub_heading' not in menu


Menu = Union[HeadedMenu, HeadlessMenu]


def menu_items(menu: Menu) -> list[Item]:
    """Return items of the menu.

    in case it's a function, the return value of the function is called.
    """
    return menu['items']() if\
        callable(menu['items']) else\
        menu['items']


def menu_title(menu: Menu) -> str:
    """Return items of the menu.

    in case it's a function, the return value of the function is called.
    """
    return menu['title']() if\
        callable(menu['title']) else\
        menu['title']


class BaseItem(TypedDict):
    """A class used to represent a menu item.

    Attributes
    ----------
    label: `str`
        The label of the item.

    color: `str` or `tuple` of `float`
        The color by its name or in rgba format as a list of floats, the list should
    contain 4 elements: red, green, blue and alpha, each being a number in the range
    [0..1].
        For example (0.5, 0, 0.5, 0.8) represents a semi transparent purple, or the
    string 'yellow' represents color yellow.

    background_color: `str` or `tuple` of `float`
        The background color by color name or in rgba format as a list of floats, the
    list should contain 4 elements red, green, blue and alpha, each being a number in
    the range [0..1].
        For example (0.5, 0, 0.5, 0.8) represents a semi transparent purple, or the
    string 'yellow' represents color yellow.

    icon: `str`
        Name of a Material Symbols icon, check here for the complete list:
    https://fonts.google.com/icons

    is_short: `bool`
        Whether the item should be rendered in short form or not. In short form only the
    icon of the item is rendered and its label is hidden.
    """

    label: str
    color: NotRequired[str | tuple[float, float, float, float]]
    background_color: NotRequired[str | tuple[float, float, float, float]]
    icon: NotRequired[str]
    is_short: NotRequired[bool]


class ActionItem(BaseItem):
    """A class used to represent an action menu item.

    Attributes
    ----------
    action: `Function`
        If provided, activating this item will call this function.
    """

    action: Callable


def is_action_item(item: Item) -> TypeGuard[ActionItem]:
    """Check whether the item is an `ActionItem` or not."""
    return 'action' in item


class ApplicationItem(BaseItem):
    """A class used to represent an application menu item.

    Attributes
    ----------
    application: `PageWidget`
        If provided, activating this item will show this widget
    """

    application: PageWidget


def is_application_item(item: Item) -> TypeGuard[ApplicationItem]:
    """Check whether the item is an `ApplicationItem` or not."""
    return 'application' in item


class SubMenuItem(BaseItem):
    """A class used to represent a sub-menu menu item.

    Attributes
    ----------
    sub_menu: `Menu`
        If provided, activating this item will open another menu, the description
        described in this field.
    """

    sub_menu: Menu


def is_sub_menu_item(item: Item) -> TypeGuard[SubMenuItem]:
    """Check whether the item is an `SubMenuItem` or not."""
    return 'sub_menu' in item


Item = Union[ActionItem, SubMenuItem, ApplicationItem]
